@@ -4,7 +4,8 @@ import datetime
 from itertools import islice, izip, chain, ifilter
 from collections import OrderedDict
 
-from sql import Table, Column, Literal, Desc, Asc, Expression, Null
+from sql import (Table, Column, Literal, Desc, Asc, Expression, Null,
+    NullsFirst, NullsLast)
 from sql.functions import CurrentTimestamp, Extract
 from sql.conditionals import Coalesce
 from sql.operators import Or, And, Operator
@@ -145,37 +146,21 @@ class ModelSQL(ModelStorage):
         for field_name, field in cls._fields.iteritems():
             if field_name == 'id':
                 continue
-            default_fun = None
-            if hasattr(field, 'set'):
-                continue
             sql_type = field.sql_type()
+            if not sql_type:
+                continue
+
+            default = None
             if field_name in cls._defaults:
-                default_fun = cls._defaults[field_name]
+                def default():
+                    default_ = cls._clean_defaults({
+                            field_name: cls._defaults[field_name](),
+                            })[field_name]
+                    return field.sql_format(default_)
 
-                def unpack_wrapper(fun):
-                    def unpack_result(*a):
-                        try:
-                            # XXX ugly hack: some default fct try
-                            # to access the non-existing table
-                            result = fun(*a)
-                        except Exception:
-                            return None
-                        clean_results = cls._clean_defaults(
-                            {field_name: result})
-                        return clean_results[field_name]
-                    return unpack_result
-                default_fun = unpack_wrapper(default_fun)
-
-            if hasattr(field, 'size') and isinstance(field.size, int):
-                field_size = field.size
-            else:
-                field_size = None
-
-            table.add_raw_column(field_name, sql_type, field.sql_format,
-                default_fun, field_size, string=field.string)
+            table.add_column(field_name, field._sql_type, default=default)
             if cls._history:
-                history_table.add_raw_column(field_name, sql_type, None,
-                    string=field.string)
+                history_table.add_column(field_name, field._sql_type)
 
             if isinstance(field, (fields.Integer, fields.Float)):
                 # migration from tryton 2.2
@@ -186,19 +171,33 @@ class ModelSQL(ModelStorage):
 
             if isinstance(field, fields.Many2One):
                 if field.model_name in ('res.user', 'res.group'):
+                    # XXX need to merge ir and res
                     ref = field.model_name.replace('.', '_')
                 else:
-                    ref = pool.get(field.model_name)._table
+                    ref_model = pool.get(field.model_name)
+                    if (issubclass(ref_model, ModelSQL)
+                            and not ref_model.table_query()):
+                        ref = ref_model._table
+                        # Create foreign key table if missing
+                        if not TableHandler.table_exist(ref):
+                            TableHandler(ref_model)
+                    else:
+                        ref = None
                 if field_name in ['create_uid', 'write_uid']:
                     # migration from 3.6
                     table.drop_fk(field_name)
-                else:
+                elif ref:
                     table.add_fk(field_name, ref, field.ondelete)
 
             table.index_action(
                 field_name, action=field.select and 'add' or 'remove')
 
             required = field.required
+            # Do not set 'NOT NULL' for Binary field as the database column
+            # will be left empty if stored in the filestore or filled later by
+            # the set method.
+            if isinstance(field, fields.Binary):
+                required = False
             table.not_null_action(
                 field_name, action=required and 'add' or 'remove')
 
@@ -228,7 +227,7 @@ class ModelSQL(ModelStorage):
                 cursor.execute(*history_table.select(history_table.id))
                 if not cursor.fetchone():
                     columns = [n for n, f in cls._fields.iteritems()
-                        if not hasattr(f, 'set')]
+                        if f.sql_type()]
                     cursor.execute(*history_table.insert(
                             [Column(history_table, c) for c in columns],
                             sql_table.select(*(Column(sql_table, c)
@@ -240,16 +239,11 @@ class ModelSQL(ModelStorage):
     def _update_history_table(cls):
         TableHandler = backend.get('TableHandler')
         if cls._history:
-            table = TableHandler(cls)
             history_table = TableHandler(cls, history=True)
-            for column_name in table._columns:
-                string = ''
-                if column_name in cls._fields:
-                    string = cls._fields[column_name].string
-                history_table.add_raw_column(column_name,
-                    (table._columns[column_name]['typname'],
-                        table._columns[column_name]['typname']),
-                    None, string=string)
+            for field_name, field in cls._fields.iteritems():
+                if not field.sql_type():
+                    continue
+                history_table.add_column(field_name, field._sql_type)
 
     @classmethod
     def _get_error_messages(cls):
@@ -264,29 +258,33 @@ class ModelSQL(ModelStorage):
         return None
 
     @classmethod
-    def __raise_integrity_error(cls, exception, values, field_names=None):
+    def __raise_integrity_error(
+            cls, exception, values, field_names=None, transaction=None):
         pool = Pool()
+        TableHandler = backend.get('TableHandler')
         if field_names is None:
             field_names = cls._fields.keys()
+        if transaction is None:
+            transaction = Transaction()
         for field_name in field_names:
             if field_name not in cls._fields:
                 continue
             field = cls._fields[field_name]
             # Check required fields
             if (field.required
-                    and not hasattr(field, 'set')
+                    and field.sql_type()
                     and field_name not in ('create_uid', 'create_date')):
                 if values.get(field_name) is None:
                     cls.raise_user_error('required_field',
                         error_args=cls._get_error_args(field_name))
             if isinstance(field, fields.Many2One) and values.get(field_name):
                 Model = pool.get(field.model_name)
-                create_records = Transaction().create_records.get(
+                create_records = transaction.create_records.get(
                     field.model_name, set())
-                delete_records = Transaction().delete_records.get(
+                delete_records = transaction.delete_records.get(
                     field.model_name, set())
                 target_records = Model.search([
-                        ('id', '=', values[field_name]),
+                        ('id', '=', field.sql_format(values[field_name])),
                         ], order=[])
                 if not ((target_records
                             or (values[field_name] in create_records))
@@ -296,10 +294,10 @@ class ModelSQL(ModelStorage):
                     cls.raise_user_error('foreign_model_missing',
                         error_args=error_args)
         for name, _, error in cls._sql_constraints:
-            if name in str(exception):
+            if TableHandler.convert_name(name) in str(exception):
                 cls.raise_user_error(error)
         for name, error in cls._sql_error_messages.iteritems():
-            if name in str(exception):
+            if TableHandler.convert_name(name) in str(exception):
                 cls.raise_user_error(error)
 
     @classmethod
@@ -335,7 +333,7 @@ class ModelSQL(ModelStorage):
         return revisions
 
     @classmethod
-    def __insert_history(cls, ids, deleted=False):
+    def _insert_history(cls, ids, deleted=False):
         transaction = Transaction()
         cursor = transaction.connection.cursor()
         if not cls._history:
@@ -354,7 +352,7 @@ class ModelSQL(ModelStorage):
                 'write_date': cls.write_date,
                 }
         for fname, field in sorted(fields.iteritems()):
-            if hasattr(field, 'set'):
+            if not field.sql_type():
                 continue
             columns.append(Column(table, fname))
             hcolumns.append(Column(history, fname))
@@ -384,7 +382,7 @@ class ModelSQL(ModelStorage):
         columns = []
         hcolumns = []
         fnames = sorted(n for n, f in cls._fields.iteritems()
-            if not hasattr(f, 'set'))
+            if f.sql_type())
         for fname in fnames:
             columns.append(Column(table, fname))
             if fname == 'write_uid':
@@ -430,9 +428,9 @@ class ModelSQL(ModelStorage):
             for sub_ids in grouped_slice(to_delete):
                 where = reduce_ids(table.id, sub_ids)
                 cursor.execute(*table.delete(where=where))
-            cls.__insert_history(to_delete, True)
+            cls._insert_history(to_delete, True)
         if to_update:
-            cls.__insert_history(to_update)
+            cls._insert_history(to_update)
 
     @classmethod
     def restore_history(cls, ids, datetime):
@@ -543,9 +541,11 @@ class ModelSQL(ModelStorage):
                         id_new = transaction.database.lastid(cursor)
                 new_ids.append(id_new)
             except DatabaseIntegrityError, exception:
+                transaction = Transaction()
                 with Transaction().new_transaction(), \
                         Transaction().set_context(_check_access=False):
-                    cls.__raise_integrity_error(exception, values)
+                    cls.__raise_integrity_error(
+                        exception, values, transaction=transaction)
                 raise
 
         domain = Rule.domain_get(cls.__name__, mode='create')
@@ -594,7 +594,7 @@ class ModelSQL(ModelStorage):
             field = cls._fields[fname]
             field.set(cls, fname, *fargs)
 
-        cls.__insert_history(new_ids)
+        cls._insert_history(new_ids)
 
         records = cls.browse(new_ids)
         for sub_records in grouped_slice(records, cache_size()):
@@ -663,7 +663,7 @@ class ModelSQL(ModelStorage):
         columns = []
         for f in fields_names + fields_related.keys() + datetime_fields:
             field = cls._fields.get(f)
-            if field and not hasattr(field, 'set'):
+            if field and field.sql_type():
                 columns.append(field.sql_column(table).as_(f))
             elif f == '_timestamp' and not table_query:
                 sql_type = fields.Char('timestamp').sql_type().base
@@ -709,22 +709,37 @@ class ModelSQL(ModelStorage):
         else:
             result = [{'id': x} for x in ids]
 
+        cachable_fields = []
         for column in columns:
             # Split the output name to remove SQLite type detection
-            field = column.output_name.split()[0]
-            if field == '_timestamp':
+            fname = column.output_name.split()[0]
+            if fname == '_timestamp':
                 continue
-            if (getattr(cls._fields[field], 'translate', False)
-                    and not hasattr(field, 'set')):
-                translations = Translation.get_ids(cls.__name__ + ',' + field,
-                    'model', Transaction().language, ids)
-                for row in result:
-                    row[field] = translations.get(row['id']) or row[field]
+            field = cls._fields[fname]
+            if not hasattr(field, 'get'):
+                if getattr(field, 'translate', False):
+                    translations = Translation.get_ids(
+                        cls.__name__ + ',' + fname, 'model',
+                        Transaction().language, ids)
+                    for row in result:
+                        row[fname] = translations.get(row['id']) or row[fname]
+                if fname != 'id':
+                    cachable_fields.append(fname)
 
         # all fields for which there is a get attribute
         getter_fields = [f for f in
             fields_names + fields_related.keys() + datetime_fields
             if f in cls._fields and hasattr(cls._fields[f], 'get')]
+
+        if getter_fields and cachable_fields:
+            cache = transaction.get_cache().setdefault(
+                cls.__name__, LRUDict(cache_size()))
+            for row in result:
+                if row['id'] not in cache:
+                    cache[row['id']] = {}
+                for fname in cachable_fields:
+                    cache[row['id']][fname] = row[fname]
+
         func_fields = {}
         for fname in getter_fields:
             field = cls._fields[fname]
@@ -926,10 +941,12 @@ class ModelSQL(ModelStorage):
                     cursor.execute(*table.update(columns, update_values,
                             where=red_sql))
                 except DatabaseIntegrityError, exception:
-                    with Transaction().new_transaction() as transaction, \
-                            transaction.set_context(_check_access=False):
-                        cls.__raise_integrity_error(exception, values,
-                            values.keys())
+                    transaction = Transaction()
+                    with Transaction().new_transaction(), \
+                            Transaction().set_context(_check_access=False):
+                        cls.__raise_integrity_error(
+                            exception, values, values.keys(),
+                            transaction=transaction)
                     raise
 
             for fname, value in values.iteritems():
@@ -950,7 +967,7 @@ class ModelSQL(ModelStorage):
             field = cls._fields[fname]
             field.set(cls, fname, *fargs)
 
-        cls.__insert_history(all_ids)
+        cls._insert_history(all_ids)
         for sub_records in grouped_slice(all_records, cache_size()):
             cls._validate(sub_records, field_names=all_field_names)
         cls.trigger_write(trigger_eligibles)
@@ -1084,13 +1101,15 @@ class ModelSQL(ModelStorage):
             try:
                 cursor.execute(*table.delete(where=red_sql))
             except DatabaseIntegrityError, exception:
+                transaction = Transaction()
                 with Transaction().new_transaction():
-                    cls.__raise_integrity_error(exception, {})
+                    cls.__raise_integrity_error(
+                        exception, {}, transaction=transaction)
                 raise
 
         Translation.delete_ids(cls.__name__, 'model', ids)
 
-        cls.__insert_history(ids, deleted=True)
+        cls._insert_history(ids, deleted=True)
 
         cls._update_mptt(tree_ids.keys(), tree_ids.values())
 
@@ -1111,14 +1130,25 @@ class ModelSQL(ModelStorage):
             'DESC': Desc,
             'ASC': Asc,
             }
+        null_ordering_types = {
+            'NULLS FIRST': NullsFirst,
+            'NULLS LAST': NullsLast,
+            None: lambda _: _
+            }
         if order is None or order is False:
             order = cls._order
         for oexpr, otype in order:
             fname, _, extra_expr = oexpr.partition('.')
             field = cls._fields[fname]
-            Order = order_types[otype.upper()]
+            otype = otype.upper()
+            try:
+                otype, null_ordering = otype.split(' ', 1)
+            except ValueError:
+                null_ordering = None
+            Order = order_types[otype]
+            NullOrdering = null_ordering_types[null_ordering]
             forder = field.convert_order(oexpr, tables, cls)
-            order_by.extend((Order(o) for o in forder))
+            order_by.extend((NullOrdering(Order(o)) for o in forder))
 
         # construct a clause for the rules :
         domain = Rule.domain_get(cls.__name__, mode='read')
@@ -1131,7 +1161,7 @@ class ModelSQL(ModelStorage):
         table = convert_from(None, tables)
 
         if count:
-            cursor.execute(*table.select(Count(Literal(1)),
+            cursor.execute(*table.select(Count(Literal('*')),
                     where=expression, limit=limit, offset=offset))
             return cursor.fetchone()[0]
         # execute the "main" query to fetch the ids we were searching for
